"""Handle the RADIUS socket
"""
from eventlet.green import socket

class RadiusSocket:
    """Handle the RADIUS socket"""
    def __init__(self, listen_ip, listen_port, server_ip, server_port):
        self.socket = None
        self.listen_ip = listen_ip
        self.listen_port = listen_port
        self.server_ip = server_ip
        self.server_port = server_port

    def setup(self):
        """Setup RADIUS Socket"""
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # pylint: disable=no-member
        self.socket.bind((self.listen_ip, self.listen_port))

    def send(self, data):
        """Sends on the radius socket
            data (bytes): what to send"""
        self.socket.sendto(data, (self.server_ip, self.server_port))

    def receive(self):
        """Receives from the radius socket"""
        return self.socket.recv(4096)




# pylint: disable=too-few-public-methods

from chewie.radius import RadiusAttributesList, RadiusAccessRequest, Radius
from chewie.radius_attributes import CallingStationId, UserName, MessageAuthenticator, EAPMessage, \
    NASPort
from chewie.ethernet_packet import EthernetPacket
from chewie.auth_8021x import Auth8021x
from chewie.eap import Eap, EapIdentity, EapMd5Challenge, EapSuccess, EapFailure, EapLegacyNak, \
    EapTTLS, EapTLS, EapPEAP, PARSERS_TYPES
from chewie.utils import MessageParseError


class EapMessage:
    src_mac = None
    message_id = None

    def __init__(self, src_mac, message_id):
        self.src_mac = src_mac
        self.message_id = message_id


class SuccessMessage(EapMessage):

    @classmethod
    def build(cls, src_mac, eap):
        return cls(src_mac, eap.packet_id)


class FailureMessage(EapMessage):

    @classmethod
    def build(cls, src_mac, eap):
        return cls(src_mac, eap.packet_id)


class IdentityMessage(EapMessage):
    def __init__(self, src_mac, message_id, code, identity):
        super().__init__(src_mac, message_id)
        self.code = code
        self.identity = identity

    @classmethod
    def build(cls, src_mac, eap):
        return cls(src_mac, eap.packet_id, eap.code, eap.identity)


class LegacyNakMessage(EapMessage):
    def __init__(self, src_mac, message_id, code, desired_auth_types):
        super().__init__(src_mac, message_id)
        self.code = code
        self.desired_auth_types = desired_auth_types

    @classmethod
    def build(cls, src_mac, eap):
        return cls(src_mac, eap.packet_id, eap.code, eap.desired_auth_types)


class Md5ChallengeMessage(EapMessage):
    def __init__(self, src_mac, message_id, code, challenge, extra_data):
        super().__init__(src_mac, message_id)
        self.code = code
        self.challenge = challenge
        self.extra_data = extra_data

    @classmethod
    def build(cls, src_mac, eap):
        return cls(src_mac, eap.packet_id, eap.code, eap.challenge, eap.extra_data)


class TlsMessageBase(EapMessage):
    """TLS and TTLS will extend this class, but TTLS cannot be same type as TLS"""
    def __init__(self, src_mac, message_id, code, flags, extra_data):
        super().__init__(src_mac, message_id)
        self.code = code
        self.flags = flags
        self.extra_data = extra_data

    @classmethod
    def build(cls, src_mac, eap):
        return cls(src_mac, eap.packet_id, eap.code, eap.flags, eap.extra_data)


class TlsMessage(TlsMessageBase):
    pass


class TtlsMessage(TlsMessageBase):
    pass


class PeapMessage(TlsMessageBase):
    pass


class EapolStartMessage(EapMessage):
    def __init__(self, src_mac):
        super().__init__(src_mac, None)

    @classmethod
    def build(cls, src_mac):
        return cls(src_mac)


class EapolLogoffMessage(EapMessage):
    def __init__(self, src_mac):
        super().__init__(src_mac, None)

    @classmethod
    def build(cls, src_mac):
        return cls(src_mac)


EAP_MESSAGES = {
    Eap.IDENTITY: IdentityMessage,
    Eap.MD5_CHALLENGE: Md5ChallengeMessage,
    Eap.LEGACY_NAK: LegacyNakMessage,
    Eap.TLS: TlsMessage,
    Eap.TTLS: TtlsMessage,
    Eap.PEAP: PeapMessage,
}


AUTH_8021X_MESSAGES = {
    0: "eap",
    1: "eapol start",
}


class MessageParser:
    @staticmethod
    def one_x_parse(data, src_mac):
        """Parses the 1x header (version and packet type) part of the packet, and the payload.
        Args:
            data:
            src_mac (MacAddress): source mac address of the data packet.
        Raises:
            MessageParseError: the data cannot be parsed."""
        auth_8021x = Auth8021x.parse(data)
        if auth_8021x.packet_type == 0:
            return MessageParser.eap_parse(auth_8021x.data, src_mac)
        elif auth_8021x.packet_type == 1:
            return EapolStartMessage.build(src_mac)
        elif auth_8021x.packet_type == 2:
            return EapolLogoffMessage.build(src_mac)
        raise MessageParseError("802.1x has bad type, expected 0: %s" % auth_8021x)

    @staticmethod
    def eap_parse(data, src_mac):
        """Parses the actual EAP payload
        Args:
            data:
            src_mac (MacAddress): source mac address of the data packet
        Raises:
            MessageParseError: the data cannot be parsed."""
        eap = Eap.parse(data)

        if isinstance(eap, tuple(PARSERS_TYPES.values())):
            return EAP_MESSAGES[eap.PACKET_TYPE].build(src_mac, eap)
        elif isinstance(eap, EapSuccess):
            return SuccessMessage.build(src_mac, eap)
        elif isinstance(eap, EapFailure):
            return FailureMessage.build(src_mac, eap)
        else:
            raise MessageParseError("Got bad Eap packet: %s" % eap)

    @staticmethod
    def ethernet_parse(packed_message):
        """Parses the ethernet header part, and payload
        Args:
            packed_message:
        Returns:
            ***Message & destination mac address.
        Raises:
            MessageParseError: the packed_message cannot be parsed."""
        ethernet_packet = EthernetPacket.parse(packed_message)
        if ethernet_packet.ethertype != 0x888e:
            raise MessageParseError("Ethernet packet with bad ethertype received: %s" %
                                    ethernet_packet)

        return MessageParser.one_x_parse(ethernet_packet.data, ethernet_packet.src_mac), \
            ethernet_packet.dst_mac

    @staticmethod
    def radius_parse(packed_message, secret, radius_lifecycle):
        """Parses a RADIUS packet
        Returns:
            RadiusPacket
        Raises:
            MessageParseError: the packed_message cannot be parsed"""
        parsed_radius = Radius.parse(packed_message, secret,
                                     radius_lifecycle=radius_lifecycle)
        return parsed_radius


class MessagePacker:
    @staticmethod
    def ethernet_pack(message, src_mac, dst_mac):
        """Packs a ethernet packet.
        Args:
            message: EAP payload
            src_mac (MacAddress):
            dst_mac (MacAddress):
        Returns:
            packed ethernet packet (bytes)
        """
        data = MessagePacker.pack(message)
        ethernet_packet = EthernetPacket(dst_mac=dst_mac, src_mac=src_mac,
                                         ethertype=0x888e, data=data)
        return ethernet_packet.pack()

    @staticmethod
    def radius_pack(eap_message, src_mac, username, radius_packet_id,
                    request_authenticator, state, secret, nas_port=None, extra_attributes=None):
        """
        Packs up a RADIUS message to send to a RADIUS Server.
        Args:
            eap_message (Message): e.g. IdentityMessage
            src_mac (MacAddress): supplicants mac address
            username (str): supplicants username
            radius_packet_id (int):
            request_authenticator (bytes):
            state (State): RADIUS State
            secret (str): RADIUS secret used between Chewie and RADIUS Server
            extra_attributes (list): list of extra RADIUS attributes to send along with the above.

        Returns:
            packed RADIUS packet (bytes)
        """
        if not extra_attributes:
            extra_attributes = []

        attr_list = []
        attr_list.append(UserName.create(username))
        attr_list.append(CallingStationId.create(str(src_mac)))

        if nas_port:
            attr_list.append(NASPort.create(nas_port))

        attr_list.extend(extra_attributes)

        attr_list.append(EAPMessage.create(eap_message))

        if state:
            attr_list.append(state)

        attr_list.append(MessageAuthenticator.create(
            bytes.fromhex("00000000000000000000000000000000")))

        attributes = RadiusAttributesList(attr_list)
        access_request = RadiusAccessRequest(radius_packet_id, request_authenticator, attributes)
        return access_request.build(secret)

    @staticmethod
    def eap_pack(message):
        """
        Pack an EAP message.
        Args:
            message (Message):

        Returns:
            version (int), packet_type (int), packed eap (bytes)
        """
        if isinstance(message, IdentityMessage):

            eap = EapIdentity(message.code, message.message_id, message.identity)
            version = 1
            packet_type = 0
            data = eap.pack()
        elif isinstance(message, LegacyNakMessage):
            eap = EapLegacyNak(message.code, message.message_id, message.desired_auth_types)
            version = 1
            packet_type = 0
            data = eap.pack()
        elif isinstance(message, Md5ChallengeMessage):
            eap = EapMd5Challenge(message.code, message.message_id,
                                  message.challenge, message.extra_data)
            version = 1
            packet_type = 0
            data = eap.pack()
        elif isinstance(message, TlsMessage):
            eap = EapTLS(message.code, message.message_id, message.flags, message.extra_data)
            version = 1
            packet_type = 0
            data = eap.pack()
        elif isinstance(message, TtlsMessage):
            eap = EapTTLS(message.code, message.message_id, message.flags, message.extra_data)
            version = 1
            packet_type = 0
            data = eap.pack()
        elif isinstance(message, PeapMessage):
            eap = EapPEAP(message.code, message.message_id, message.flags, message.extra_data)
            version = 1
            packet_type = 0
            data = eap.pack()
        elif isinstance(message, SuccessMessage):
            eap = EapSuccess(message.message_id)
            version = 1
            packet_type = 0
            data = eap.pack()
        elif isinstance(message, FailureMessage):
            eap = EapFailure(message.message_id)
            version = 1
            packet_type = 0
            data = eap.pack()
        elif isinstance(message, EapolStartMessage):
            version = 1
            packet_type = 1
            data = b""
        elif isinstance(message, EapolLogoffMessage):
            version = 1
            packet_type = 2
            data = b""
        else:
            raise ValueError("Cannot pack message: %s" % message)
        return version, packet_type, data

    @staticmethod
    def pack(message):
        """
        packs the EAPOL
        Args:
            message (Message): EAP message

        Returns:
            Packed EAPOL packet (bytes)
        """
        version, packet_type, data = MessagePacker.eap_pack(message)
        auth_8021x = Auth8021x(version=version, packet_type=packet_type, data=data)
        return auth_8021x.pack()



"""Mock TimerScheduler
"""

class FakeTimerJob:
    """Behaves like TimerJob"""
    def __init__(self, function, args, timeout):
        self.function = function
        self.args = args
        self.timeout = timeout
        self.is_cancelled = False

    def cancel(self):
        """Clones TimerJob.cancel()"""
        self.is_cancelled = True

    def cancelled(self):
        """Clones TimerJob.cancelled()"""
        return self.is_cancelled

    def run(self):
        """Runs job"""
        if not self.is_cancelled:
            self.function(*self.args)

class FakeTimerScheduler:
    """Behaves like TimerScheduler"""
    def __init__(self):
        self.jobs = []

    def call_later(self, timeout, func, *args):
        """Clones TimerScheduler.call_later()"""
        if not args:
            args = []

        job = FakeTimerJob(func, args, timeout)

        self.jobs.append(job)

        return job

    def run_jobs(self):
        """Runs jobs in order of timeout"""
        while self.jobs:
            self.jobs.sort(key=lambda x: x.timeout)
            job = self.jobs.pop(0)
            job.run()

    def run(self):
        """Clones TimerScheduler.run()"""
        pass



"""RADIUS Packets"""
import copy
import binascii
import hashlib
import hmac
import struct

from chewie.radius_attributes import ATTRIBUTE_TYPES, Attribute, MessageAuthenticator
from chewie.radius_datatypes import Concat
from chewie.utils import MessageParseError

RADIUS_HEADER_LENGTH = 1 + 1 + 2 + 16

PACKET_TYPE_PARSERS = {}


class InvalidResponseAuthenticatorError(Exception):
    """To be used when the ResponseAuthenticator hashes
     (received in packet, and calculated) do not match."""
    pass


class InvalidMessageAuthenticatorError(Exception):
    """To be used when the Message-Authenticator hashes
     (received in packet, and calculated) do not match.
    Received packets that throw this error should be 'silently dropped' (logging is fine)."""
    pass


class Radius:
    """Radius packet interface which will determin the correct RadiusPacket child class to use"""
    ACCESS_REQUEST = 1
    ACCESS_ACCEPT = 2
    ACCESS_REJECT = 3
    ACCOUNTING_REQUEST = 4
    ACCOUNTING_RESPONSE = 5
    ACCESS_CHALLENGE = 11
    STATUS_SERVER = 12
    STATUS_CLIENT = 13

    @staticmethod
    def parse(packed_message, secret, radius_lifecycle=None):
        """
        Args:
            packed_message:
            secret (str): Shared sceret between chewie and RADIUS server.
            radius_lifecycle: RadiusLifecycle object
        Returns:
            RadiusPacket - RadiusAccessChallenge/RadiusAccessRequest/
                            RadiusAccessAccept/RadiusAccessFailure
        Raises:
            MessageParseError: if packed_message cannot be parsed
        """
        try:
            code, packet_id, length, authenticator = struct.unpack("!BBH16s",
                                                                   packed_message[:RADIUS_HEADER_LENGTH])
        except struct.error as exception:
            raise MessageParseError('Unable to unpack first 20 bytes of RADIUS header') \
                from exception

        if code in PACKET_TYPE_PARSERS.keys():
            radius_packet = PACKET_TYPE_PARSERS[code](packet_id, authenticator,
                                                      RadiusAttributesList.parse(
                                                          packed_message[RADIUS_HEADER_LENGTH:]))
            if code == Radius.ACCESS_REQUEST:
                request_authenticator = authenticator
            else:
                try:
                    request_authenticator = radius_lifecycle.packet_id_to_request_authenticator[
                        packet_id]
                except KeyError as exception:
                    raise MessageParseError('Unknown RAIDUS packet_id: %s' % packet_id, ) \
                        from exception
            try:
                return radius_packet.validate_packet(secret,
                                                     request_authenticator=request_authenticator,
                                                     code=code)
            except (InvalidMessageAuthenticatorError,
                    InvalidResponseAuthenticatorError) as exception:
                raise MessageParseError("Unable to parse Radius packet") \
                    from exception
        raise MessageParseError("Unable to parse radius code: %d" % code)

    def pack(self):
        pass


def register_packet_type_parser(cls):
    PACKET_TYPE_PARSERS[cls.CODE] = cls.parse
    return cls


class RadiusPacket(Radius):
    """super class for different radius packets"""
    CODE = None
    packed = None

    def __init__(self, packet_id, authenticator, attributes):
        self.packet_id = packet_id
        self.authenticator = authenticator
        self.attributes = attributes

    @classmethod
    def parse(cls, packet_id, request_authenticator, attributes):
        return cls(packet_id, request_authenticator, attributes)

    def pack(self):
        header = struct.pack("!BBH16s", self.CODE, self.packet_id,
                             RADIUS_HEADER_LENGTH + self.attributes.__len__(),
                             self.authenticator)
        packed_attributes = self.attributes.pack()
        self.packed = bytearray(header + packed_attributes)
        return self.packed

    def build(self, secret=None):
        """Only call this once, or else the MessageAuthenticator will not be zeros,
         resulting in the wrong hash
         Args:
             secret (str): Shared sceret between chewie and RADIUS server.
        Returns:
            packed packet (bytes)"""
        if not self.packed:
            self.pack()
        try:
            position = self.attributes.indexof(MessageAuthenticator.DESCRIPTION) + \
                       RADIUS_HEADER_LENGTH + Attribute.HEADER_SIZE
        except ValueError as e:
            print(e.message)
            return self.packed

        if secret:
            message_authenticator = bytearray(hmac.new(secret.encode(), self.packed, 'md5')
                                              .digest())

            for i in range(16):
                self.packed[i + position] = message_authenticator[i]
        return self.packed

    def validate_packet(self, secret, request_authenticator=None, code=None):
        """Calculates the Response Authenticator (in Radius Header) and
        MessageAuthenticator (a Radius Attribute) hashes and compares with what was provided.
        Args:
            code (int): The RADIUS Code (e.g. Access-Challenge)
            secret (str): secret shared between RADIUS and chewie.
            request_authenticator (): the original request authenticator for this
             packet (which is a response)
        Raises:
            ValueError: if secret is None or empty string.
            InvalidResponseAuthenticatorError: if Response Authenticator does not match calculated.
            InvalidMessageAuthenticatorError: if MessageAuthenticator does not match calculated.
        """
        # Copy this packet so we can modify the 'Authenticator' and 'Message-Authenticator'
        radius_packet = copy.deepcopy(self)

        if not secret:
            raise ValueError("secret cannot be None for hashing")

        self.validate_response_authenticator(radius_packet, request_authenticator, secret, code)

        self.validate_message_authenticator(radius_packet, secret, request_authenticator)
        return self

    @staticmethod
    def validate_response_authenticator(radius_packet, request_authenticator, secret, code):
        if request_authenticator and code in [Radius.ACCESS_REJECT,
                                              Radius.ACCESS_ACCEPT,
                                              Radius.ACCESS_CHALLENGE]:
            response_authenticator = radius_packet.authenticator
            radius_packet.authenticator = request_authenticator
            radius_packet.pack()
            calculated_response_authenticator = hashlib.md5(radius_packet.packed +
                                                            bytearray(secret, 'utf-8')).digest()
            if calculated_response_authenticator != response_authenticator:
                raise InvalidResponseAuthenticatorError(
                    "Original ResponseAuthenticator: '%s', does not match calculated: '%s' %s" % (
                        response_authenticator,
                        calculated_response_authenticator,
                        binascii.hexlify(radius_packet.packed)))

    @staticmethod
    def validate_message_authenticator(radius_packet, secret, request_authenticator):
        message_authenticator = radius_packet.attributes.find(MessageAuthenticator.DESCRIPTION)
        if message_authenticator:

            radius_packet.authenticator = request_authenticator

            original_ma = message_authenticator.data_type.bytes_data
            # Replace the Original Message Authenticator
            message_authenticator.data_type.bytes_data = bytes.fromhex(
                "00000000000000000000000000000000")

            radius_packet.pack()

            # calculate new hash message authenticator
            new_ma = hmac.new(secret.encode(), radius_packet.packed, 'md5').digest()

            # compare old and new message authenticator
            if original_ma != new_ma:
                raise InvalidMessageAuthenticatorError(
                    "Original Message-Authenticator: '%s', does not match calculated: '%s'" %
                    (binascii.hexlify(original_ma), binascii.hexlify(new_ma)))


@register_packet_type_parser
class RadiusAccessRequest(RadiusPacket):
    CODE = Radius.ACCESS_REQUEST


@register_packet_type_parser
class RadiusAccessAccept(RadiusPacket):
    CODE = Radius.ACCESS_ACCEPT


@register_packet_type_parser
class RadiusAccessReject(RadiusPacket):
    CODE = Radius.ACCESS_REJECT


@register_packet_type_parser
class RadiusAccessChallenge(RadiusPacket):
    CODE = Radius.ACCESS_CHALLENGE


class RadiusAttributesList:
    """Container class for the Radius Attribute Value Pairs"""

    def __init__(self, attributes):
        self.attributes = attributes

    @classmethod
    def parse(cls, attributes_data):
        """

        Args:
            attributes_data:

        Returns:
            RadiusAttributeList
        Raises:
            MessageParseError: if unable to parse an attribute's data.
        """
        attributes = []
        attributes_to_concat = {}
        cls.extract_attributes(attributes_data, attributes, attributes_to_concat)

        attributes = cls.merge_concat_attributes(attributes, attributes_to_concat)

        return cls(attributes)

    @classmethod
    def merge_concat_attributes(cls, attributes, attributes_to_concat):
        """
        Removes concat attributes for attributes list, and inserts a single new master attribute
        for all concat attributes of the same type (e.g. EAPMessage, EAPMessage, = 1 EAPMessage)
        Args:
            attributes (list):
            attributes_to_concat (dict): attribute - position.
        Returns:
            attributes (list)
        Raises:
            MessageParseError: RadiusAttribute.parse will raise error
            if it cannot parse the attribute's data
        """
        # Join Attributes that's datatype is Concat into one attribute.
        concatenated_attributes = []
        for value, list_ in attributes_to_concat.items():
            concatenated_data = b""
            for d, i in list_:
                concatenated_data += d.data_type.bytes_data
            concatenated_attributes.append(tuple((ATTRIBUTE_TYPES[value].parse(concatenated_data),
                                                  i)))
        # Remove old Attributes that were concatenated.
        for ca, _ in concatenated_attributes:
            attributes = [x for x in attributes if x.TYPE != ca.TYPE]

        # need to put them back in the same position.
        for ca, i in concatenated_attributes:
            attributes.insert(i, ca)

        return attributes

    @classmethod
    def extract_attributes(cls, attributes_data, attributes, attributes_to_concat):
        """
        Extracts Radius Attributes from a packed payload.
        Keeps track of attribute ordering.
        Args:
            attributes_data (): data to extract from (input).
            attributes: attributes extracted (output variable).
            attributes_to_concat (dict): (output variable).
        Raises:
            MessageParseError: RadiusAttribute.parse will raise error
            if it cannot parse the attribute's data
        """
        total_length = len(attributes_data)
        pos = 0
        index = -1
        last_attribute = -1
        while pos < total_length:
            try:
                type_, attr_length = struct.unpack("!BB",
                                                   attributes_data[pos:pos + Attribute.HEADER_SIZE])
            except struct.error as exception:
                raise MessageParseError('Unable to unpack first 2 bytes of attribute header') \
                    from exception
            data = attributes_data[pos + Attribute.HEADER_SIZE: pos + attr_length]
            pos += attr_length

            packed_value = data[:attr_length - Attribute.HEADER_SIZE]
            try:
                attribute = ATTRIBUTE_TYPES[type_].parse(packed_value)
            except KeyError as exception:
                raise MessageParseError('Cannot find parser for RADIUS attribute %s' %
                                        type_) from exception
            # keep track of where the concated AVP should be in the attributes list.
            # required so the hashing gives correct hash.
            if attribute.DATA_TYPE != Concat or last_attribute != attribute.TYPE:
                index += 1
            last_attribute = attribute.TYPE

            if attribute.DATA_TYPE.DATA_TYPE_VALUE == Concat.DATA_TYPE_VALUE:
                if attribute.TYPE not in attributes_to_concat:
                    attributes_to_concat[attribute.TYPE] = []
                attributes_to_concat[attribute.TYPE].append((attribute, index))

            attributes.append(attribute)

    def find(self, item):
        """Find first attribute that has the matching description
        Args:
            item (str): description of attribute to find
        Returns:
            attribute or None if not found"""
        for attr in self.attributes:
            if item == attr.DESCRIPTION:
                return attr
        return None

    def indexof(self, item):
        """Finds the position (number of bytes) that item is at in list.
        Args:
            item (str): description of attribute to find index of.
        Returns:
            int - number of bytes to item.
        Raises:
            ValueErrpr: if cannot find item
        """
        i = 0
        for attr in self.attributes:
            if item == attr.DESCRIPTION:
                break
            i += attr.full_length()
        else:
            raise ValueError("Cannot find item: %s in attributes list" % item)
        return i

    def __len__(self):
        total = 0
        for attr in self.attributes:
            total = total + attr.full_length()
        return total

    def pack(self):
        packed_attributes = bytes()
        for attr in self.attributes:
            packed_attributes += attr.pack()
        return packed_attributes

    def to_dict(self):
        ret = {}
        for a in self.attributes:
            ret[a.DESCRIPTION] = a.data_type.data()
        return ret



"""A placeholder object for RADIUS logic extracted from Chewie"""

import os
import struct

from chewie.message_parser import MessagePacker
from chewie.radius_attributes import EAPMessage, State, CalledStationId, NASIdentifier, NASPortType
from chewie.event import EventRadiusMessageReceived


def port_id_to_int(port_id):
    """"Convert a port_id str '00:00:00:aa:00:01 to integer'"""
    dp, port_half_1, port_half_2 = str(port_id).split(':')[3:]
    port = port_half_1 + port_half_2
    return int.from_bytes(struct.pack('!HH', int(dp, 16), int(port, 16)), 'big')  # pytype: disable=attribute-error


class RadiusLifecycle:
    """A placeholder object for RADIUS logic extracted from Chewie"""
    def __init__(self, radius_secret, server_id, logger):
        self.radius_secret = radius_secret
        self.server_id = server_id
        self.logger = logger

        self.next_radius_id = 0
        self.extra_radius_request_attributes = self.prepare_extra_radius_attributes()

        self.packet_id_to_mac = {}  # radius_packet_id: mac
        self.packet_id_to_request_authenticator = {}

    def process_outbound(self, radius_output_bits):
        """Placeholder method extracted from Chewie.send_radius_messages()"""
        eap_message = radius_output_bits.message
        src_mac = radius_output_bits.src_mac
        username = radius_output_bits.identity
        state = radius_output_bits.state
        port_id = radius_output_bits.port_mac

        self.logger.info("got eap to send to radius.. mac: %s %s, username: %s",
                         type(src_mac), src_mac, username)
        state_dict = None
        if state:
            state_dict = state.__dict__
        self.logger.info("Sending to RADIUS eap message %s with state %s",
                         eap_message.__dict__, state_dict)
        radius_packet_id = self.get_next_radius_packet_id()
        self.packet_id_to_mac[radius_packet_id] = {'src_mac': src_mac, 'port_id': port_id}
        # message is eap. needs to be wrapped into a radius packet.
        request_authenticator = self.generate_request_authenticator()
        self.packet_id_to_request_authenticator[radius_packet_id] = request_authenticator
        return MessagePacker.radius_pack(eap_message, src_mac, username,
                                         radius_packet_id, request_authenticator, state,
                                         self.radius_secret,
                                         port_id_to_int(port_id),
                                         self.extra_radius_request_attributes)

    def build_event_radius_message_received(self, radius):
        """Build a EventRadiusMessageReceived from a radius message"""
        eap_msg_attribute = radius.attributes.find(EAPMessage.DESCRIPTION)
        eap_msg = eap_msg_attribute.data_type.data()
        state = radius.attributes.find(State.DESCRIPTION)
        self.logger.info("radius EAP: %s", eap_msg)
        return EventRadiusMessageReceived(eap_msg, state, radius.attributes.to_dict())

    def generate_request_authenticator(self):
        """Workaround until we get this extracted for easy mocking"""
        return os.urandom(16)

    def get_next_radius_packet_id(self):
        """Calulate the next RADIUS Packet ID
        Returns:
            int
        """
        radius_id = self.next_radius_id
        self.next_radius_id = (self.next_radius_id + 1) % 256

        return radius_id

    def prepare_extra_radius_attributes(self):
        """Create RADIUS Attirbutes to be sent with every RADIUS request"""
        attr_list = [CalledStationId.create(self.server_id),
                     NASPortType.create(15),
                     NASIdentifier.create(self.server_id)]
        return attr_list



"""Unittests for chewie/chewie.py"""

import logging
import random
import sys
import tempfile
import unittest
from unittest.mock import patch

import eventlet
from eventlet.queue import Queue

from chewie.chewie import Chewie
from chewie.eap_state_machine import FullEAPStateMachine
from chewie.mac_address import MacAddress
from helpers import FakeTimerScheduler


FROM_SUPPLICANT = Queue()
TO_SUPPLICANT = Queue()
FROM_RADIUS = Queue()
TO_RADIUS = Queue()


def patch_things(func):
    """decorator to mock patch socket operations and random number generators"""
    @patch('chewie.chewie.EapSocket', FakeEapSocket)
    @patch('chewie.chewie.RadiusSocket', FakeRadiusSocket)
    @patch('chewie.chewie.RadiusLifecycle.generate_request_authenticator', urandom_helper)
    @patch('chewie.chewie.FullEAPStateMachine.next_id', next_id)
    def wrapper_patch(self):
        func(self)

    return wrapper_patch


def setup_generators(_supplicant_replies=None, _radius_replies=None):
    """decorator to setup the packets for the mocked socket (queues) to send"""
    def decorator_setup_gen(func):
        def wrapper_setup_gen(self):
            global SUPPLICANT_REPLY_GENERATOR  # pylint: disable=global-statement
            global RADIUS_REPLY_GENERATOR  # pylint: disable=global-statement
            global URANDOM_GENERATOR  # pylint: disable=global-statement

            SUPPLICANT_REPLY_GENERATOR = supplicant_replies_gen(_supplicant_replies)
            RADIUS_REPLY_GENERATOR = radius_replies_gen(_radius_replies)
            URANDOM_GENERATOR = urandom()
            func(self)
        return wrapper_setup_gen
    return decorator_setup_gen


def supplicant_replies_gen(replies):
    """generator for packets supplicant sends"""
    for reply in replies:
        yield reply


def radius_replies_gen(replies):
    """generator for packets radius sends"""
    for reply in replies:
        yield reply


def urandom():
    """generator for urandom"""
    _list = [b'\x87\xf5[\xa71\xeeOA;}\\t\xde\xd7.=',
             b'\xf7\xe0\xaf\xc7Q!\xa2\xa9\xa3\x8d\xf7\xc6\x85\xa8k\x06']
    for random_bytes in _list:
        yield random_bytes


URANDOM_GENERATOR = None  # urandom()


def urandom_helper(size):  # pylint: disable=unused-argument
    """helper for urandom_generator"""
    return next(URANDOM_GENERATOR)


SUPPLICANT_REPLY_GENERATOR = None  # supplicant_replies()
RADIUS_REPLY_GENERATOR = None  # radius_replies()


class FakeEapSocket:
    def __init__(self, _interface_name):
        # TODO inject queues in constructor instead of using globals
        pass

    def setup(self):
        pass

    def receive(self):  # pylint: disable=unused-argument
        global FROM_SUPPLICANT

        print('mocked eap_receive')
        got = FROM_SUPPLICANT.get()
        return got

    def send(self, data=None):  # pylint: disable=unused-argument
        global TO_SUPPLICANT
        global FROM_SUPPLICANT
        global SUPPLICANT_REPLY_GENERATOR

        print('mocked eap_send')
        if data:
            TO_SUPPLICANT.put_nowait(data)
        try:
            next_reply = next(SUPPLICANT_REPLY_GENERATOR)
        except StopIteration:
            return
        if next_reply:
            FROM_SUPPLICANT.put_nowait(next_reply)


class FakeRadiusSocket:
    def __init__(self, _listen_ip, _listen_port, _server_ip, _server_port):
        # TODO inject queues in constructor instead of using globals
        pass

    def setup(self):
        pass

    def receive(self):  # pylint: disable=unused-argument
        global FROM_RADIUS

        print('mocked radius_receive')
        got = FROM_RADIUS.get()
        print('got RADIUS', got)
        return got

    def send(self, data):  # pylint: disable=unused-argument
        global TO_RADIUS
        global FROM_RADIUS
        global RADIUS_REPLY_GENERATOR

        print('mocked radius_send')
        TO_RADIUS.put_nowait(data)
        try:
            next_reply = next(RADIUS_REPLY_GENERATOR)
        except StopIteration:
            return
        if next_reply:
            FROM_RADIUS.put_nowait(next_reply)


def do_nothing(chewie):  # pylint: disable=unused-argument
    """Mock function that does nothing.
     Typically used on socket opening/configuration operations"""
    pass


def next_id(eap_state_machine):  # pylint: disable=invalid-name
    """mocked FullEAPStateMachine.nextId"""
    if eap_state_machine.current_id is None:
        return 116
    _id = eap_state_machine.current_id + 1
    if _id > 255:
        return random.randint(0, 200)
    return _id


def auth_handler(client_mac, port_id_mac):  # pylint: disable=unused-argument
    """dummy handler for successful authentications"""
    print('Successful auth from MAC %s on port: %s' % (str(client_mac), str(port_id_mac)))


def failure_handler(client_mac, port_id_mac):  # pylint: disable=unused-argument
    """dummy handler for failed authentications"""
    print('failure from MAC %s on port: %s' % (str(client_mac), str(port_id_mac)))


def logoff_handler(client_mac, port_id_mac):  # pylint: disable=unused-argument
    """dummy handler for logoffs"""
    print('logoff from MAC %s on port: %s' % (str(client_mac), str(port_id_mac)))


class ChewieTestCase(unittest.TestCase):
    """Main chewie.py test class"""

    no_radius_replies = []

    header = "0000000000010242ac17006f888e"
    sup_replies_success = [bytes.fromhex(header + "01000009027400090175736572"),
                           bytes.fromhex(
                               header + "010000160275001604103abcadc86714b2d75d09dd7ff53edf6b")]

    radius_replies_success = [bytes.fromhex(
        "0b000050066300262c8ee8a33f43ad4e837e63c54f180175001604101a16a3baa37a0238f33384f6c11067425012a3bf83bea0eeb69645088527dc491eed18126aa866456add628e3a55a4737872cad6"),
                              bytes.fromhex(
                                  "02010032ec22830bb2fbde37f635e91690410b334f060375000450129a08d246ec3da2f371ec4ff16eed9310010675736572")]

    sup_replies_logoff = [bytes.fromhex(header + "01000009027400090175736572"),
                          bytes.fromhex(
                              header + "010000160275001604103abcadc86714b2d75d09dd7ff53edf6b"),
                          bytes.fromhex("0000000000010242ac17006f888e01020000")]

    # packet id (0x84 is incorrect)
    sup_replies_failure_message_id = [bytes.fromhex(header + "01000009028400090175736572"),
                                      bytes.fromhex(header + "01000009029400090175736572"),
                                      bytes.fromhex(header + "01000009026400090175736572"),
                                      bytes.fromhex(header + "01000009025400090175736572")]

    # the first response has correct code, second is wrong and will be dropped by radius
    sup_replies_failure2_response_code = [bytes.fromhex(header + "01000009027400090175736572"),
                                          bytes.fromhex(header + "01000009037400090175736572")]

    def setUp(self):
        logger = logging.getLogger()
        logger.level = logging.DEBUG
        self.log_file = tempfile.NamedTemporaryFile()

        logger.addHandler(logging.FileHandler(self.log_file.name))
        logger.addHandler(logging.StreamHandler(sys.stdout))

        self.chewie = Chewie('lo', logger,
                             auth_handler, failure_handler, logoff_handler,
                             '127.0.0.1', 1812, 'SECRET',
                             '44:44:44:44:44:44')
        self.fake_scheduler = FakeTimerScheduler()
        self.chewie.timer_scheduler = self.fake_scheduler

        global FROM_SUPPLICANT  # pylint: disable=global-statement
        global TO_SUPPLICANT  # pylint: disable=global-statement
        global FROM_RADIUS  # pylint: disable=global-statement
        global TO_RADIUS  # pylint: disable=global-statement

        FROM_SUPPLICANT = Queue()
        TO_SUPPLICANT = Queue()
        FROM_RADIUS = Queue()
        TO_RADIUS = Queue()

    def tearDown(self):
        self.chewie.shutdown()

    def test_get_state_machine(self):
        """Tests Chewie.get_state_machine()"""
        self.assertEqual(len(self.chewie.state_machines), 0)
        # creates the state_machine if it doesn't exist
        state_machine = self.chewie.get_state_machine('12:34:56:78:9a:bc',  # pylint: disable=invalid-name
                                                      '00:00:00:00:00:01')

        self.assertEqual(len(self.chewie.state_machines), 1)

        self.assertIs(state_machine, self.chewie.get_state_machine('12:34:56:78:9a:bc',
                                                                   '00:00:00:00:00:01'))

        self.assertIsNot(state_machine, self.chewie.get_state_machine('12:34:56:78:9a:bc',
                                                                      '00:00:00:00:00:02'))
        self.assertIsNot(state_machine, self.chewie.get_state_machine('ab:cd:ef:12:34:56',
                                                                      '00:00:00:00:00:01'))

        # 2 ports
        self.assertEqual(len(self.chewie.state_machines), 2)
        # port 1 has 2 macs
        self.assertEqual(len(self.chewie.state_machines['00:00:00:00:00:01']), 2)
        # port 2 has 1 mac
        self.assertEqual(len(self.chewie.state_machines['00:00:00:00:00:02']), 1)

    def test_get_state_machine_by_packet_id(self):
        """Tests Chewie.get_state_machine_by_packet_id()"""
        self.chewie.radius_lifecycle.packet_id_to_mac[56] = {'src_mac': '12:34:56:78:9a:bc',
                                                             'port_id': '00:00:00:00:00:01'}
        state_machine = self.chewie.get_state_machine('12:34:56:78:9a:bc',  # pylint: disable=invalid-name
                                                      '00:00:00:00:00:01')

        self.assertIs(self.chewie.get_state_machine_from_radius_packet_id(56),
                      state_machine)
        with self.assertRaises(KeyError):
            self.chewie.get_state_machine_from_radius_packet_id(20)

    @patch_things
    @setup_generators(sup_replies_success, radius_replies_success)
    def test_success_dot1x(self):
        """Test success api"""
        FROM_SUPPLICANT.put_nowait(bytes.fromhex("0000000000010242ac17006f888e01010000"))
        
        pool = eventlet.GreenPool()
        pool.spawn(self.chewie.run)

        eventlet.sleep(1)

        self.assertEqual(
            self.chewie.get_state_machine('02:42:ac:17:00:6f',
                                          '00:00:00:00:00:01').state,
            FullEAPStateMachine.SUCCESS2)

    def test_port_status_changes(self):
        """test port status api"""
        # TODO what can actually be checked here?
        # the state machine tests already check the statemachine
        # could check that the preemptive identity request packet is sent. (once implemented)
        # for now just check api works under python version.

        self.chewie.port_down("00:00:00:00:00:01")

        self.chewie.port_up("00:00:00:00:00:01")

        self.chewie.port_down("00:00:00:00:00:01")

    @patch_things
    @setup_generators(sup_replies_logoff, radius_replies_success)
    def test_logoff_dot1x(self):
        """Test logoff"""
        self.chewie.get_state_machine(MacAddress.from_string('02:42:ac:17:00:6f'),
                                      MacAddress.from_string('00:00:00:00:00:01'))
        FROM_SUPPLICANT.put_nowait(bytes.fromhex("0000000000010242ac17006f888e01010000"))

        pool = eventlet.GreenPool()
        pool.spawn(self.chewie.run)

        eventlet.sleep(0.1)

        self.assertEqual(
            self.chewie.get_state_machine('02:42:ac:17:00:6f',
                                          '00:00:00:00:00:01').state,
            FullEAPStateMachine.LOGOFF2)

    @patch_things
    @setup_generators(sup_replies_failure_message_id, no_radius_replies)
    def test_failure_message_id_dot1x(self):
        """Test incorrect message id results in timeout_failure"""
        # TODO not convinced this is transitioning through the correct states.
        # (should be discarding all packets)
        # But end result is correct (both packets sent/received, and end state)self.chewie.get_state_machine(MacAddress.from_string('02:42:ac:17:00:6f'),
        self.chewie.get_state_machine(MacAddress.from_string('02:42:ac:17:00:6f'),
                                      MacAddress.from_string(
                                          '00:00:00:00:00:01')).DEFAULT_TIMEOUT = 0.5

        FROM_SUPPLICANT.put_nowait(bytes.fromhex("0000000000010242ac17006f888e01010000"))

        pool = eventlet.GreenPool()
        pool.spawn(self.chewie.run)

        while not self.fake_scheduler.jobs:
            eventlet.sleep(0.1)
        self.fake_scheduler.run_jobs()

        self.assertEqual(
            self.chewie.get_state_machine('02:42:ac:17:00:6f',
                                          '00:00:00:00:00:01').state,
            FullEAPStateMachine.TIMEOUT_FAILURE)


    @patch_things
    @setup_generators(sup_replies_failure2_response_code, no_radius_replies)
    def test_failure2_resp_code_dot1x(self):
        """Test incorrect eap.code results in timeout_failure2. RADIUS Server drops it.
        It is up to the supplicant to send another request - this supplicant doesnt"""
        self.chewie.get_state_machine(MacAddress.from_string('02:42:ac:17:00:6f'),
                                      MacAddress.from_string(
                                          '00:00:00:00:00:01')).DEFAULT_TIMEOUT = 0.5

        FROM_SUPPLICANT.put_nowait(bytes.fromhex("0000000000010242ac17006f888e01010000"))

        pool = eventlet.GreenPool()
        pool.spawn(self.chewie.run)

        while not self.fake_scheduler.jobs:
            eventlet.sleep(0.1)
        self.fake_scheduler.run_jobs()

        self.assertEqual(
            self.chewie.get_state_machine('02:42:ac:17:00:6f',
                                          '00:00:00:00:00:01').state,
            FullEAPStateMachine.TIMEOUT_FAILURE2)



import logging
import sys

from chewie.chewie import Chewie


def get_logger(name, log_level=logging.DEBUG):
    logger = logging.getLogger(name)
    if not logger.handlers:
        logger.setLevel(log_level)
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(log_level)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    return logger


def auth_handler(address, group_address):
    print("Authed address %s on port %s" % (str(address), str(group_address)))


def failure_handler(address, group_address):
    print("failure of address %s on port %s" % (str(address), str(group_address)))


def logoff_handler(address, group_address):
    print("logoff of address %s on port %s" % (str(address), str(group_address)))


logger = get_logger("CHEWIE")
logger.info('starting chewieeeee.')

chewie = Chewie("eth1", logger, auth_handler, failure_handler, logoff_handler, radius_server_ip="172.24.0.113", radius_server_secret="SECRET")
chewie.run()



"""MAC address helper"""

import re

# pylint: disable=too-few-public-methods


_MAC_REGEX = re.compile(r'(?:[0-9a-f]{1,2}:){5}[0-9a-f]{1,2}\Z', re.IGNORECASE)


class MacAddress:
    """Class for comparing mac addresses"""
    def __init__(self, address):
        self.address = address

    @classmethod
    def from_string(cls, address_string):
        """Create a MacAddress from a string.

        Args:
            address_string (str): Colon-delimited MAC address.

        Returns:
            MacAddress

        Raises:
            ValueError: If address_string is invalid.
        """
        if not _MAC_REGEX.match(address_string):
            raise ValueError("'%s' does not appear to be a MAC address" % address_string)
        address = bytes(int(x, 16) for x in address_string.split(':'))
        return cls(address)

    def __str__(self):
        address_string = ":".join("%02x" % x for x in self.address)
        return address_string

    def __eq__(self, other):
        return self.address == other.address

    def __hash__(self):
        return hash(self.address)

    def __repr__(self):
        return "%s.from_string(\"%s\")" % (self.__class__.__name__, self.__str__())



"""Radius Attributes"""
# TODO if attributes have requirements e.g. length must be above minimum, can enforce that here.
# TODO could we auto generate this from the radius-types-2.csv available from iana.org?

import struct

from chewie.radius_datatypes import Concat, Enum, Integer, String, Text, Vsa


ATTRIBUTE_TYPES = {}


class Attribute():
    """Parent class for the Attributes."""

    TYPE = None  # e.g. 1
    DATA_TYPE = None  # e.g. Text
    DESCRIPTION = None  # e.g. "User-Name"

    HEADER_SIZE = 1 + 1

    def __init__(self, data_type):
        self.data_type = data_type

    @classmethod
    def create(cls, data):
        """Factory method.
        Args:
            data: object of python type (int, str, bytes, ...)
        Returns:
            Attribute subclass.
        """
        return cls(cls.DATA_TYPE(raw_data=data))  # pylint: disable=not-callable

    @classmethod
    def parse(cls, packed_value):
        """
        Args:
            packed_value (bytes): pre-packed value
        Returns:
            Attribute subclass.
        Raises:
            MessageParseError: if unable to parse the packed_value into the appropriate datatype.
        """
        return cls(cls.DATA_TYPE.parse(packed_value))

    def pack(self):
        """
        Returns:
            packed attribute (including header) bytes
        """
        tl = struct.pack("!BB", self.TYPE, self.full_length())
        v = self.data_type.pack(self.TYPE)
        return tl + v

    def full_length(self):
        """
        Returns:
            length (including header).
        """
        return self.data_type.full_length()


def register_attribute_type(cls):
    """Decoratot to register RADIUS attribute types"""
    ATTRIBUTE_TYPES[cls.TYPE] = cls
    return cls


@register_attribute_type
class UserName(Attribute):
    """User-Name https://tools.ietf.org/html/rfc2865#section-5.1"""
    TYPE = 1
    DATA_TYPE = Text
    DESCRIPTION = "User-Name"


@register_attribute_type
class NASPort(Attribute):
    """Service-Type https://tools.ietf.org/html/rfc2865#section-5.5"""
    TYPE = 5
    DATA_TYPE = Integer
    DESCRIPTION = "NAS-Port"


@register_attribute_type
class ServiceType(Attribute):
    """Service-Type https://tools.ietf.org/html/rfc2865#section-5.6"""
    TYPE = 6
    DATA_TYPE = Enum
    DESCRIPTION = "Service-Type"


@register_attribute_type
class FramedMTU(Attribute):
    """Framed-MTU https://tools.ietf.org/html/rfc2865#section-5.12"""
    TYPE = 12
    DATA_TYPE = Integer
    DESCRIPTION = "Framed-MTU"


@register_attribute_type
class ReplyMessage(Attribute):
    """Reply-Message https://tools.ietf.org/html/rfc2865#section-5.18"""
    TYPE = 18
    DATA_TYPE = Text
    DESCRIPTION = "Reply-Message"


@register_attribute_type
class State(Attribute):
    """State https://tools.ietf.org/html/rfc2865#section-5.24"""
    TYPE = 24
    DATA_TYPE = String
    DESCRIPTION = "State"

    # TODO length >= 3 https://tools.ietf.org/html/rfc2865#section-5.24


@register_attribute_type
class VendorSpecific(Attribute):
    """Vendor-Specific https://tools.ietf.org/html/rfc2865#section-5.26"""
    TYPE = 26
    DATA_TYPE = Vsa
    DESCRIPTION = "Vendor-Specific"


@register_attribute_type
class SessionTimeout(Attribute):
    """Vendor-Specific https://tools.ietf.org/html/rfc2865#section-5.27"""
    TYPE = 27
    DATA_TYPE = Integer
    DESCRIPTION = "Session-Timeout"


@register_attribute_type
class CalledStationId(Attribute):
    """Called-Station-Id https://tools.ietf.org/html/rfc2865#section-5.30"""
    TYPE = 30
    DATA_TYPE = Text
    DESCRIPTION = "Called-Station-Id"


@register_attribute_type
class CallingStationId(Attribute):
    """Calling-Station-Id https://tools.ietf.org/html/rfc2865#section-5.31"""
    TYPE = 31
    DATA_TYPE = Text
    DESCRIPTION = "Calling-Station-Id"


@register_attribute_type
class NASIdentifier(Attribute):
    """Calling-Station-Id https://tools.ietf.org/html/rfc2865#section-5.32"""
    TYPE = 32
    DATA_TYPE = Text
    DESCRIPTION = "NAS-Identifier"


@register_attribute_type
class AcctSessionId(Attribute):
    """Acct-Session-id (RADIUS Accounting) https://tools.ietf.org/html/rfc2866#section-5.5"""
    TYPE = 44
    DATA_TYPE = Text
    DESCRIPTION = "Acct-Session-Id"


@register_attribute_type
class NASPortType(Attribute):
    """NAS-Port-Type https://tools.ietf.org/html/rfc2865#section-5.41"""
    TYPE = 61
    DATA_TYPE = Enum
    DESCRIPTION = "NAS-Port-Type"


@register_attribute_type
class ConnectInfo(Attribute):
    """ConnectInfo (RADIUS Extensions) https://tools.ietf.org/html/rfc2869#section-5.11"""
    TYPE = 77
    DATA_TYPE = Text
    DESCRIPTION = "Connect-Info"


@register_attribute_type
class EAPMessage(Attribute):
    """EAP-Message (RADIUS Extensions) https://tools.ietf.org/html/rfc2869#section-5.13"""
    TYPE = 79
    DATA_TYPE = Concat
    DESCRIPTION = "EAP-Message"

    def pack(self):
        """Concat types need to override AttributeType.pack().
        as Concat.pack() may return multiple packed AVP (each with their own length)"""
        return self.data_type.pack(self.TYPE)


@register_attribute_type
class MessageAuthenticator(Attribute):
    """Message-Authenticator (RADIUS Extensions) https://tools.ietf.org/html/rfc2869#section-5.14"""
    TYPE = 80
    DATA_TYPE = String
    DESCRIPTION = "Message-Authenticator"



"""Homebrew Event scheduler, as sched.scheduler was not working outside of unittests"""
import heapq
import time

import eventlet


class TimerJob:
    """Represents a job for TimerScheduler, same api as asyncio.TimerHandle"""

    expiry_time = 0
    is_cancelled = False
    func = None
    args = None

    def __init__(self, expiry_time, func, args):
        self.expiry_time = expiry_time
        self.func = func
        self.args = args

    def cancel(self):
        """Cancel the callback."""
        self.is_cancelled = True

    def cancelled(self):
        """
        Returns:
            True if callback was cancelled
        """
        return self.is_cancelled

    def when(self):
        """
        Returns:
            scheduled callback time as float seconds
        """
        return self.expiry_time


class TimerScheduler:
    """wraps a heapbased queue with a similar api to asyncio.loop"""

    def __init__(self, logger, sleep=None):
        self.logger = logger
        self.timer_heap = []

        self.sleep = eventlet.sleep
        if sleep:
            self.sleep = sleep

    def call_later(self, timeout, func, *args):
        """Scheduler callback.

        Args:
            timeout: number of seconds to delay executing func
            func: function to execute
            *args: arguments for func

        Returns:
            tuple (expiry_time, dict) - can be used for cancelling the job
        """
        if not args:
            args = []
        expiry_time = time.time() + timeout

        job = TimerJob(expiry_time, func, args)
        heapq.heappush(self.timer_heap, (expiry_time, job))
        return job

    def run(self):
        """Main loop. should run forever"""
        while True:
            try:
                if self.timer_heap:
                    if self.timer_heap[0][0] < time.time():
                        _, job = heapq.heappop(self.timer_heap)
                        if not job.cancelled():
                            self.logger.info('running job %s %s', job.func.__name__, job.args)
                            job.func(*job.args)
                        else:
                            self.logger.info('job %s has been cancelled', job.func.__name__)
                    else:
                        self.sleep(1)
                else:
                    self.sleep(1)
            except Exception as e:
                self.logger.exception(e)




# pylint: disable=missing-docstring

import unittest

from chewie.mac_address import MacAddress


class MacAddressTestCase(unittest.TestCase):
    """Test MacAddress class."""

    def test_from_string_valid(self):
        """Test valid mac address input."""
        examples = {
            '01:02:03:04:05:06': b'\x01\x02\x03\x04\x05\x06',
            '01:02:03:04:05:7': b'\x01\x02\x03\x04\x05\x07',
            '0a:0b:0c:0d:0e:0f': b'\x0a\x0b\x0c\x0d\x0e\x0f',
            '0a:0b:0c:0d:0e:f': b'\x0a\x0b\x0c\x0d\x0e\x0f',
            '1:2:3:4:5:6': b'\x01\x02\x03\x04\x05\x06',
            'a:b:c:D:E:F': b'\x0a\x0b\x0c\x0d\x0e\x0f'
        }
        for key, data in examples.items():
            addr = MacAddress.from_string(key)
            self.assertEqual(addr.address, data)

    def test_from_string_invalid(self):
        """Test invalid mac address input."""
        examples = [
            '01:02:03:04:05',
            '01::03:04:05:06',
            '01:02:03:04:05: 6',
            '01:02:03:04:05:6\n',
            ' 1:02:03:04:05:06',
            '01:02:03:04:05:06:07',
            '0:0:0:0:0:007',
            '0a:0b:0c:0d:0e:0g',
            '0a:0b:0c:0d:0e:fff',
            '01-02-03-04-05-06',
            '0001:0002:0003',
            '000001:000002'
        ]
        for key in examples:
            with self.assertRaisesRegex(ValueError, r'%s.+MAC address' % key):
                MacAddress.from_string(key)

    def test_hash_and_equals(self):
        """Test mac address __hash__ and __eq__."""
        addr1 = MacAddress.from_string('01:02:03:04:05:06')
        addr2 = MacAddress(b'\x01\x02\x03\x04\x05\x06')
        self.assertEqual(addr1, addr2)
        self.assertEqual(hash(addr1), hash(addr2))

    def test_repr(self):
        """Test mac address __repr__."""
        value = repr(MacAddress.from_string('01:02:03:04:05:06'))
        self.assertEqual(value, 'MacAddress.from_string("01:02:03:04:05:06")')



"""Entry point for 802.1X speaker.
"""

from eventlet import sleep, GreenPool
from eventlet.queue import Queue

from chewie import timer_scheduler
from chewie.eap_socket import EapSocket
from chewie.radius_socket import RadiusSocket
from chewie.eap_state_machine import FullEAPStateMachine
from chewie.radius_lifecycle import RadiusLifecycle
from chewie.message_parser import MessageParser, MessagePacker
from chewie.event import EventMessageReceived, EventPortStatusChange
from chewie.utils import get_logger, MessageParseError


def unpack_byte_string(byte_string):
    """unpacks a byte string"""
    return "".join("%02x" % x for x in byte_string)


class Chewie:
    """Facilitates EAP supplicant and RADIUS server communication"""
    RADIUS_UDP_PORT = 1812

    def __init__(self, interface_name, logger=None,
                 auth_handler=None, failure_handler=None, logoff_handler=None,
                 radius_server_ip=None, radius_server_port=None, radius_server_secret=None,
                 chewie_id=None):
        self.interface_name = interface_name
        self.logger = get_logger(logger.name + "." + Chewie.__name__)
        self.auth_handler = auth_handler
        self.failure_handler = failure_handler
        self.logoff_handler = logoff_handler

        self.radius_server_ip = radius_server_ip
        self.radius_secret = radius_server_secret
        self.radius_server_port = self.RADIUS_UDP_PORT
        if radius_server_port:
            self.radius_server_port = radius_server_port
        self.radius_listen_ip = "0.0.0.0"
        self.radius_listen_port = 0

        self.chewie_id = "44-44-44-44-44-44:"  # used by the RADIUS Attribute
                                               # 'Called-Station' in Access-Request
        if chewie_id:
            self.chewie_id = chewie_id

        self.state_machines = {}  # mac: state_machine

        self.eap_output_messages = Queue()
        self.radius_output_messages = Queue()

        self.radius_lifecycle = RadiusLifecycle(self.radius_secret, self.chewie_id, self.logger)
        self.timer_scheduler = timer_scheduler.TimerScheduler(self.logger)

        self.eap_socket = None
        self.pool = None
        self.eventlets = None
        self.radius_socket = None
        self.interface_index = None

        self.eventlets = []

    def run(self):
        """setup chewie and start socket eventlet threads"""
        self.logger.info("Starting")
        self.setup_eap_socket()
        self.setup_radius_socket()
        self.start_threads_and_wait()

    def running(self):
        """Used to nicely exit the event loops"""
        return True

    def shutdown(self):
        """kill eventlets and quit"""
        for eventlet in self.eventlets:
            eventlet.kill()

    def start_threads_and_wait(self):
        """Start the thread and wait until they complete (hopefully never)"""
        self.pool = GreenPool()

        self.eventlets.append(self.pool.spawn(self.send_eap_messages))
        self.eventlets.append(self.pool.spawn(self.receive_eap_messages))

        self.eventlets.append(self.pool.spawn(self.send_radius_messages))
        self.eventlets.append(self.pool.spawn(self.receive_radius_messages))

        self.eventlets.append(self.pool.spawn(self.timer_scheduler.run))

        self.pool.waitall()

    def auth_success(self, src_mac, port_id):
        """authentication shim between faucet and chewie
        Args:
            src_mac (MacAddress): the mac of the successful supplicant
            port_id (MacAddress): the 'mac' identifier of what switch port the success is on"""
        if self.auth_handler:
            self.auth_handler(src_mac, port_id)

    def auth_failure(self, src_mac, port_id):
        """failure shim between faucet and chewie
        Args:
            src_mac (MacAddress): the mac of the failed supplicant
            port_id (MacAddress): the 'mac' identifier of what switch port
             the failure is on"""
        if self.failure_handler:
            self.failure_handler(src_mac, port_id)

    def auth_logoff(self, src_mac, port_id):
        """logoff shim between faucet and chewie
        Args:
            src_mac (MacAddress): the mac of the logoff supplicant
            port_id (MacAddress): the 'mac' identifier of what switch port
             the logoff is on"""
        if self.logoff_handler:
            self.logoff_handler(src_mac, port_id)

    def port_down(self, port_id):
        """
        should be called by faucet when port has gone down.
        Args:
            port_id (str): id of port.
        """
        # all chewie needs to do is change its internal state.
        # faucet will remove the acls by itself.
        self.set_port_status(port_id, False)

    def port_up(self, port_id):
        """
        should be called by faucet when port has come up
        Args:
            port_id (str): id of port.
        """
        self.set_port_status(port_id, True)
        # TODO send preemptive identity request.

    def set_port_status(self, port_id, status):
        port_id_str = str(port_id)
        if port_id_str not in self.state_machines:
            self.state_machines[port_id_str] = {}

        for src_mac, state_machine in self.state_machines[port_id_str].items():
            event = EventPortStatusChange(status)
            state_machine.event(event)

    def setup_eap_socket(self):
        self.eap_socket = EapSocket(self.interface_name)
        self.eap_socket.setup()

    def setup_radius_socket(self):
        self.radius_socket = RadiusSocket(self.radius_listen_ip,
                                          self.radius_listen_port,
                                          self.radius_server_ip,
                                          self.radius_server_port)
        self.radius_socket.setup()
        self.logger.info("Radius Listening on %s:%d" % (self.radius_listen_ip,
                                                        self.radius_listen_port))

    def send_eap_messages(self):
        """send eap messages to supplicant forever."""
        while self.running():
            sleep(0)
            eap_queue_message = self.eap_output_messages.get()
            self.logger.info("Sending message %s from %s to %s" %
                             (eap_queue_message.message, str(eap_queue_message.port_mac),
                              str(eap_queue_message.src_mac)))
            self.eap_socket.send(MessagePacker.ethernet_pack(eap_queue_message.message,
                                                             eap_queue_message.port_mac,
                                                             eap_queue_message.src_mac))

    def receive_eap_messages(self):
        """receive eap messages from supplicant forever."""
        while self.running():
            sleep(0)
            self.logger.info("waiting for eap.")
            packed_message = self.eap_socket.receive()
            self.logger.info("Received packed_message: %s", str(packed_message))
            try:
                eap, dst_mac = MessageParser.ethernet_parse(packed_message)
            except MessageParseError as exception:
                self.logger.info(
                    "MessageParser.ethernet_parse threw exception.\n"
                    " packed_message: '%s'.\n"
                    " exception: '%s'.",
                    packed_message,
                    exception)
                continue
            self.send_eap_to_state_machine(eap, dst_mac)

    def send_eap_to_state_machine(self, eap, dst_mac):
        """sends an eap message to the state machine"""
        self.logger.info("eap EAP(): %s", eap)
        state_machine = self.get_state_machine(eap.src_mac, dst_mac)
        event = EventMessageReceived(eap, dst_mac)
        state_machine.event(event)

    def send_radius_messages(self):
        """send RADIUS messages to RADIUS Server forever."""
        while self.running():
            sleep(0)
            radius_output_bits = self.radius_output_messages.get()
            packed_message = self.radius_lifecycle.process_outbound(radius_output_bits)
            self.radius_socket.send(packed_message)
            self.logger.info("sent radius message.")

    def receive_radius_messages(self):
        """receive RADIUS messages from RADIUS server forever."""
        while self.running():
            sleep(0)
            self.logger.info("waiting for radius.")
            packed_message = self.radius_socket.receive()
            try:
                radius = MessageParser.radius_parse(packed_message, self.radius_secret,
                                                    self.radius_lifecycle)
            except MessageParseError as exception:
                self.logger.info(
                    "MessageParser.radius_parse threw exception.\n"
                    " packed_message: '%s'.\n"
                    " exception: '%s'.",
                    packed_message,
                    exception)
                continue
            self.send_radius_to_state_machine(radius)
            self.logger.info("Received RADIUS message: %s", radius)

    def send_radius_to_state_machine(self, radius):
        """sends a radius message to the state machine"""
        event = self.radius_lifecycle.build_event_radius_message_received(radius)
        state_machine = self.get_state_machine_from_radius_packet_id(radius.packet_id)
        state_machine.event(event)

    def get_state_machine_from_radius_packet_id(self, packet_id):
        """Gets a FullEAPStateMachine from the RADIUS message packet_id
        Args:
            packet_id (int): id of the received RADIUS message
        Returns:
            FullEAPStateMachine
        """
        return self.get_state_machine(**self.radius_lifecycle.packet_id_to_mac[packet_id])

    def get_state_machine(self, src_mac, port_id):
        """Gets or creates if it does not already exist an FullEAPStateMachine for the src_mac.
        Args:
            src_mac (MacAddress): who's to get.
            port_id (MacAddress): ID of the port where the src_mac is.

        Returns:
            FullEAPStateMachine
        """
        port_id_str = str(port_id)
        src_mac_str = str(src_mac)
        port_state_machines = self.state_machines.get(port_id_str, None)
        if port_state_machines is None:
            self.state_machines[port_id_str] = {}
        state_machine = self.state_machines[port_id_str].get(src_mac_str, None)
        if not state_machine:
            state_machine = FullEAPStateMachine(self.eap_output_messages, self.radius_output_messages, src_mac,
                                                self.timer_scheduler, self.auth_success,
                                                self.auth_failure, self.auth_logoff, self.logger.name)
            state_machine.eap_restart = True
            # TODO what if port is not actually enabled, but then how did they auth?
            state_machine.port_enabled = True
            self.state_machines[port_id_str][src_mac_str] = state_machine
        return state_machine




# pylint: disable=missing-docstring

import unittest
from chewie.eap import Eap, EapIdentity, EapMd5Challenge, EapSuccess, EapFailure


class EapTestCase(unittest.TestCase):
    def test_eap_identity_parses(self):
        packed_message = bytes.fromhex("0101000501")
        message = Eap.parse(packed_message)
        self.assertEqual(message.packet_id, 1)
        self.assertEqual(message.identity, "")

    def test_eap_md5_challenge_parses(self):
        packed_message = bytes.fromhex(
            "0201002204103a535f0ee8c6b34fe714aa7dad9a0e154a6f686e2e4d63477569726b")
        message = Eap.parse(packed_message)
        self.assertEqual(message.packet_id, 1)
        self.assertEqual(message.challenge, bytes.fromhex("3a535f0ee8c6b34fe714aa7dad9a0e15"))
        self.assertEqual(message.extra_data, b"John.McGuirk")

    def test_eap_identity_packs(self):
        expected_packed_message = bytes.fromhex("0101000501")
        eap = EapIdentity(Eap.REQUEST, 1, "")
        packed_message = eap.pack()
        self.assertEqual(expected_packed_message, packed_message)

    def test_eap_md5_challenge_packs(self):
        expected_packed_message = bytes.fromhex(
            "0201002204103a535f0ee8c6b34fe714aa7dad9a0e154a6f686e2e4d63477569726b")
        eap = EapMd5Challenge(Eap.RESPONSE,
                              1,
                              bytes.fromhex("3a535f0ee8c6b34fe714aa7dad9a0e15"),
                              b"John.McGuirk")
        packed_message = eap.pack()
        self.assertEqual(expected_packed_message, packed_message)

    def test_eap_success_parses(self):
        packed_message = bytes.fromhex("03010004")
        message = Eap.parse(packed_message)
        self.assertEqual(message.packet_id, 1)
        self.assertTrue(isinstance(message, EapSuccess))

    def test_eap_failure_parses(self):
        packed_message = bytes.fromhex("04010004")
        message = Eap.parse(packed_message)
        self.assertEqual(message.packet_id, 1)
        self.assertTrue(isinstance(message, EapFailure))



"""Radius Attribute Datatypes"""
import math
import struct

import chewie.message_parser
from chewie.utils import MessageParseError


class DataType():
    """Parent datatype class, subclass should provide implementation for abstractmethods.
    May """
    DATA_TYPE_VALUE = None
    AVP_HEADER_LEN = 1 + 1
    MAX_DATA_LENGTH = 253
    MIN_DATA_LENGTH = 1

    bytes_data = None  # bytes version of raw_data

    def parse(self, packed_value):
        """"""
        return

    def pack(self, attribute_type):
        """"""
        return

    def data(self):
        """Subclass should override this as needed.
        Returns:
             The python type (int, str, bytes) of the bytes_data.
         This will perform any decoding as required instead of using the unprocessed bytes_data.
        """
        return self.bytes_data

    def data_length(self):
        """
        Returns:
             length of the data field, and not total length of the attribute (including the
         type and length).
        If total is required use full_length.
        """
        return 0

    def full_length(self):
        """
        Returns:
            Length of the whole field include the header (type and length)
        """
        return self.data_length() + self.AVP_HEADER_LEN

    @classmethod
    def is_valid_length(cls, packed_value):
        length = len(packed_value)
        if length < cls.MIN_DATA_LENGTH \
                or length > cls.MAX_DATA_LENGTH \
                or len(packed_value) > cls.MAX_DATA_LENGTH \
                or length != len(packed_value):
            raise ValueError("RADIUS data type '%s' length must be: %d <= actual_length(%d) <= %d"
                             ""
                             % (cls.__name__, cls.MIN_DATA_LENGTH, length, cls.MAX_DATA_LENGTH))


class Integer(DataType):
    DATA_TYPE_VALUE = 1
    MAX_DATA_LENGTH = 4
    MIN_DATA_LENGTH = 4

    def __init__(self, bytes_data=None, raw_data=None):
        if raw_data:
            try:
                bytes_data = raw_data.to_bytes(self.MAX_DATA_LENGTH, "big")
            except OverflowError as exception:
                raise ValueError("Integer must be >= 0  and <= 2^32-1, was %d" %
                                                raw_data)
        self.bytes_data = bytes_data

    @classmethod
    def parse(cls, packed_value):

        try:
            cls.is_valid_length(packed_value)
            return cls(bytes_data=struct.unpack("!4s", packed_value)[0])
        except (ValueError, struct.error)as exception:
            raise MessageParseError("%s unable to unpack." % cls.__name__) from exception

    def pack(self, attribute_type):
        return struct.pack("!4s", self.bytes_data)

    def data(self):
        return int.from_bytes(self.bytes_data, 'big') # pytype: disable=attribute-error

    def data_length(self):
        return 4


class Enum(DataType):
    DATA_TYPE_VALUE = 2
    MAX_DATA_LENGTH = 4
    MIN_DATA_LENGTH = 4

    def __init__(self, bytes_data=None, raw_data=None):
        if raw_data:
            try:
                bytes_data = raw_data.to_bytes(self.MAX_DATA_LENGTH, "big")
            except OverflowError:
                raise ValueError("Integer must be >= 0  and <= 2^32-1, was %d" % raw_data)
        self.bytes_data = bytes_data

    @classmethod
    def parse(cls, packed_value):
        try:
            cls.is_valid_length(packed_value)
            return cls(bytes_data=struct.unpack("!4s", packed_value)[0])
        except (ValueError, struct.error)as exception:
            raise MessageParseError("%s unable to unpack." % cls.__name__) from exception

    def pack(self, attribute_type):
        return struct.pack("!4s", self.bytes_data)

    def data(self):
        return int.from_bytes(self.bytes_data, 'big') # pytype: disable=attribute-error

    def data_length(self):
        return 4


class Text(DataType):
    DATA_TYPE_VALUE = 4

    def __init__(self, bytes_data=None, raw_data=None):
        if raw_data is not None:
            bytes_data = raw_data.encode()
            self.is_valid_length(bytes_data)
        self.bytes_data = bytes_data

    @classmethod
    def parse(cls, packed_value):
        try:
            cls.is_valid_length(packed_value)
            return cls(struct.unpack("!%ds" % len(packed_value), packed_value)[0])
        except (ValueError, struct.error) as exception:
            raise MessageParseError("%s unable to unpack." % cls.__name__) from exception

    def pack(self, attribute_type):
        return struct.pack("!%ds" % len(self.bytes_data), self.bytes_data)

    def data(self):
        return self.bytes_data.decode("UTF-8")

    def data_length(self):
        return len(self.bytes_data)


class String(DataType):
    # how is this different from Text?? - text is utf8
    DATA_TYPE_VALUE = 5

    def __init__(self, bytes_data=None, raw_data=None):
        if raw_data is not None:
            if isinstance(raw_data, bytes):
                bytes_data = raw_data
            else:
                bytes_data = raw_data.encode()
            self.is_valid_length(bytes_data)
        self.bytes_data = bytes_data

    @classmethod
    def parse(cls, packed_value):
        try:
            cls.is_valid_length(packed_value)
            return cls(struct.unpack("!%ds" % len(packed_value), packed_value)[0])
        except (ValueError, struct.error) as exception:
            raise MessageParseError("%s unable to unpack." % cls.__name__) from exception

    def pack(self, attribute_type):
        return struct.pack("!%ds" % len(self.bytes_data), self.bytes_data)

    def data_length(self):
        return len(self.bytes_data)


class Concat(DataType):
    """AttributeTypes that use Concat must override their pack()"""

    DATA_TYPE_VALUE = 6

    def __init__(self, bytes_data=None, raw_data=None):
        if raw_data:
            if isinstance(raw_data, chewie.message_parser.EapMessage):
                bytes_data = chewie.message_parser.MessagePacker.eap_pack(raw_data)[2]
            else:
                bytes_data = bytes.fromhex(raw_data)
            # self.is_valid_length(data)
        self.bytes_data = bytes_data

    @classmethod
    def parse(cls, packed_value):
        # TODO how do we want to do valid length checking here?
        #
        # Parsing is (generally) for packets coming from the radius server.
        # Packing is (generally) for packets going to the radius server.
        #
        # Therefore we error out if length is too long (you are not allowed to have AVP that are too long)
        try:
            return cls(struct.unpack("!%ds" % len(packed_value), packed_value)[0])
        except struct.error as exception:
            raise MessageParseError("%s unable to unpack." % cls.__name__) from exception

    def pack(self, attribute_type):

        def chunks(data):
            length = self.MAX_DATA_LENGTH
            list_length = len(data)
            return_chunks = []
            for i in range(0, list_length, self.MAX_DATA_LENGTH):
                if i + self.MAX_DATA_LENGTH > list_length:
                    length = list_length % self.MAX_DATA_LENGTH

                chunk = data[i:i + length]
                chunk_length = len(chunk)
                packed = struct.pack("!BB%ds" % chunk_length, attribute_type,
                                     chunk_length + self.AVP_HEADER_LEN,
                                     chunk)
                return_chunks.append(packed)
            return return_chunks

        packed = b"".join(chunks(self.bytes_data))
        return packed

    def data(self):
        return chewie.message_parser.MessageParser.eap_parse(self.bytes_data, None)

    def full_length(self):
        return self.AVP_HEADER_LEN * \
               (math.ceil(len(self.bytes_data) / self.MAX_DATA_LENGTH + 1))\
               + len(self.bytes_data) - self.AVP_HEADER_LEN

    def data_length(self):
        return len(self.bytes_data)


class Vsa(DataType):

    DATA_TYPE_VALUE = 14
    VENDOR_ID_LEN = 4
    MIN_DATA_LENGTH = 5

    def __init__(self, bytes_data=None, raw_data=None):
        if raw_data:
            bytes_data = raw_data
            self.is_valid_length(bytes_data)
        self.bytes_data = bytes_data

    @classmethod
    def parse(cls, packed_value):
        # TODO Vsa.parse does not currently separate the vendor-id from the vsa-data
        # we could do that at some point (e.g. if we wanted to use Vendor-Specific)
        try:
            cls.is_valid_length(packed_value)
            return cls(struct.unpack("!%ds" % len(packed_value), packed_value)[0])
        except (ValueError, struct.error) as exception:
            raise MessageParseError("%s unable to unpack." % cls.__name__) from exception

    def pack(self, attribute_type):
        return struct.pack("!%ds" % (self.data_length()), self.bytes_data)

    def data_length(self):
        return len(self.bytes_data)



"""Unittests for chewie/chewie.py"""

from collections import namedtuple

import unittest
from unittest.mock import patch, Mock

from chewie.chewie import Chewie
from chewie.event import EventMessageReceived
from chewie.utils import EapQueueMessage


def return_if(expected, return_value):
    """allows us to do expect-this-return-that style mocking"""
    def inner_function(*args):
        """workaround to effectively give us an anonymous function"""
        if args == expected:
            return return_value
        raise Exception("Expected %s but got %s" % (expected, args))

    return inner_function

FakeLogger = namedtuple('FakeLogger', ('name',)) # pylint: disable=invalid-name
FakeEapMessage = namedtuple('FakeEapMessage', ('src_mac',)) # pylint: disable=invalid-name

class ChewieWithMocksTestCase(unittest.TestCase):
    """Main chewie.py test class"""

    def setUp(self):
        self.chewie = Chewie('lo', FakeLogger('logger name'),
                             None, None, None,
                             '127.0.0.1', 1812, 'SECRET',
                             '44:44:44:44:44:44')

    @patch("chewie.chewie.Chewie.running", Mock(side_effect=[True, False]))
    @patch("chewie.chewie.MessageParser.ethernet_parse")
    @patch("chewie.chewie.FullEAPStateMachine")
    @patch("chewie.chewie.sleep", Mock())
    def test_eap_packet_in_goes_to_new_state_machine(self, state_machine, ethernet_parse): #pylint: disable=invalid-name
        """test EAP packet creates a new state machine and is sent on"""
        self.chewie.eap_socket = Mock(**{'receive.return_value': 'message from socket'})
        ethernet_parse.side_effect = return_if(
            ('message from socket',),
            (FakeEapMessage('fake src mac'), 'fake dst mac')
            )
        self.chewie.receive_eap_messages()
        state_machine().event.assert_called_with(
            EventMessageReceived(FakeEapMessage('fake src mac'), 'fake dst mac')
            )

    @patch("chewie.chewie.Chewie.running", Mock(side_effect=[True, False]))
    @patch("chewie.chewie.MessagePacker.ethernet_pack")
    @patch("chewie.chewie.sleep", Mock())
    def test_eap_output_packet_gets_packed_and_sent(self, ethernet_pack): #pylint: disable=invalid-name
        """test EAP packet creates a new state machine and is sent on"""
        self.chewie.eap_socket = Mock()
        ethernet_pack.return_value = "packed ethernet"
        self.chewie.eap_output_messages.put_nowait(
            EapQueueMessage("output eap message", "src mac", "port mac"))
        self.chewie.send_eap_messages()
        self.chewie.eap_socket.send.assert_called_with("packed ethernet")

    @patch("chewie.chewie.Chewie.running", Mock(side_effect=[True, False]))
    @patch("chewie.chewie.MessageParser.radius_parse")
    @patch("chewie.chewie.Chewie.get_state_machine_from_radius_packet_id")
    @patch("chewie.chewie.sleep", Mock())
    def test_radius_packet_in_goes_to_state_machine(self, state_machine, radius_parse): #pylint: disable=invalid-name
        """test radius packet goes to a state machine"""
        # note that the state machine has to exist already - if not then we blow up
        fake_radius = namedtuple('Radius', ('packet_id',))('fake packet id')
        self.chewie.radius_socket = Mock(**{'receive.return_value': 'message from socket'})
        self.chewie.radius_lifecycle = Mock(**{'build_event_radius_message_received.side_effect':
            return_if(
                (fake_radius,),
                'fake event'
            )})
        radius_parse.side_effect = return_if(
            ('message from socket', 'SECRET', self.chewie.radius_lifecycle),
            fake_radius
            )
        # not checking args as we can't mock the callback
        self.chewie.receive_radius_messages()
        state_machine().event.assert_called_with(
            'fake event'
            )

    @patch("chewie.chewie.Chewie.running", Mock(side_effect=[True, False]))
    @patch("chewie.chewie.sleep", Mock())
    def test_radius_output_packet_gets_packed_and_sent(self): #pylint: disable=invalid-name
        """test EAP packet creates a new state machine and is sent on"""
        self.chewie.radius_socket = Mock()

        self.chewie.radius_output_messages.put_nowait('fake radius output bits')
        self.chewie.radius_lifecycle = Mock(**{'process_outbound.side_effect':
            return_if(
                ('fake radius output bits',),
                'packed radius'
            )})
        self.chewie.send_radius_messages()
        self.chewie.radius_socket.send.assert_called_with("packed radius")
